@@ -1,10 +1,10 @@
 package Net::DNS::Packet;
 
 #
-# $Id: Packet.pm 1282 2014-10-27 09:45:19Z willem $
+# $Id: Packet.pm 1523 2016-12-29 16:53:52Z willem $
 #
 use vars qw($VERSION);
-$VERSION = (qw$LastChangedRevision: 1282 $)[1];
+$VERSION = (qw$LastChangedRevision: 1523 $)[1];
 
 
 =head1 NAME
@@ -39,9 +39,6 @@ BEGIN {
 	require Net::DNS::RR;
 }
 
-use constant OLDDNSSEC => Net::DNS::RR->COMPATIBLE;
-my @dummy_header = OLDDNSSEC ? ( header => {} ) : ();
-
 
 =head1 METHODS
 
@@ -66,16 +63,15 @@ sub new {
 	my $class = shift;
 
 	my $self = bless {
+		status	   => 0,
 		question   => [],
 		answer	   => [],
 		authority  => [],
 		additional => [],
-		@dummy_header		## Net::DNS::SEC 0.17 compatible
 		}, $class;
 
 	$self->{question} = [Net::DNS::Question->new(@_)] if scalar @_;
 
-	$self->header->rd(1);
 	return $self;
 }
 
@@ -111,14 +107,14 @@ or corrupted during transmission.
 use constant HEADER_LENGTH => length pack 'n6', (0) x 6;
 
 sub decode {
-	my $class = shift;
+	my $class = shift;					# uncoverable pod
 	my $data  = shift;
 	my $debug = shift || 0;
 
 	my $offset = 0;
 	my $self;
 	eval {
-		local $SIG{__WARN__} = sub { die @_ };
+		local $SIG{__DIE__};
 		die 'corrupt wire-format data' if length($$data) < HEADER_LENGTH;
 
 		# header section
@@ -134,8 +130,7 @@ sub decode {
 			answer	   => [],
 			authority  => [],
 			additional => [],
-			answersize => length $$data,
-			@dummy_header	## Net::DNS::SEC 0.17 compatible
+			answersize => length $$data
 			}, $class;
 
 		# question/zone section
@@ -165,9 +160,10 @@ sub decode {
 		return $self;
 	};
 
-	if ( $debug && $self ) {
-		local $@;
-		$self->print;
+	if ($debug) {
+		local $@ = $@;
+		print $@ if $@;
+		$self->print if $self;
 	}
 
 	return wantarray ? ( $self, $offset ) : $self;
@@ -187,25 +183,21 @@ Truncation may be specified using a non-zero optional size argument.
 =cut
 
 sub data {
-	my ( $self, $size ) = @_;
-	$self->truncate($size) if $size;			# temp fix for RT#91306
 	&encode;
 }
 
 sub encode {
-	my $self = shift;
+	my ( $self, $size ) = @_;				# uncoverable pod
 
-	my $ident = $self->header->id;				# packet header
+	my $edns = $self->edns;					# EDNS support
+	my @addl = grep !$_->isa('Net::DNS::RR::OPT'), @{$self->{additional}};
+	$self->{additional} = [@addl, $edns] if $edns->_specified;
 
-	for ( my $edns = $self->edns ) {			# EDNS support
-		my @xopt = grep !$_->isa('Net::DNS::RR::OPT'), @{$self->{additional}};
-		unshift( @xopt, $edns ) if $edns->defined;
-		$self->{additional} = \@xopt;
-	}
+	return $self->truncate($size) if $size;
 
 	my @part = qw(question answer authority additional);
 	my @size = map scalar( @{$self->{$_}} ), @part;
-	my $data = pack 'n6', $ident, $self->{status}, @size;
+	my $data = pack 'n6', $self->header->id, $self->{status}, @size;
 	$self->{count} = [];
 
 	my $hash = {};						# packet body
@@ -232,13 +224,14 @@ sub header {
 }
 
 
-=head2 EDNS extended header
+=head2 edns
 
     $edns    = $packet->edns;
     $version = $edns->version;
-    $size    = $edns->size;
+    $UDPsize = $edns->size;
 
-Auxiliary function edns() provides access to EDNS extensions.
+Auxiliary function which provides access to the EDNS protocol
+extension OPT RR.
 
 =cut
 
@@ -246,7 +239,8 @@ sub edns {
 	my $self = shift;
 	my $link = \$self->{xedns};
 	($$link) = grep $_->isa(qw(Net::DNS::RR::OPT)), @{$self->{additional}} unless $$link;
-	$$link ||= new Net::DNS::RR( type => 'OPT' );
+	$$link = new Net::DNS::RR( type => 'OPT' ) unless $$link;
+	return $$link;
 }
 
 
@@ -266,7 +260,7 @@ sub reply {
 	my $query  = shift;
 	my $UDPmax = shift;
 	my $qheadr = $query->header;
-	die 'erroneous qr flag in query packet' if $qheadr->qr;
+	croak 'erroneous qr flag in query packet' if $qheadr->qr;
 
 	my $reply  = new Net::DNS::Packet();
 	my $header = $reply->header;
@@ -276,12 +270,16 @@ sub reply {
 	my @question = $query->question;
 	$reply->{question} = [@question];
 
-	$header->rcode('FORMERR');				# failure to provide RCODE is sinful!
+	$header->rcode('FORMERR');				# no RCODE considered sinful!
 
 	$header->rd( $qheadr->rd );				# copy these flags into reply
 	$header->cd( $qheadr->cd );
 
-	$reply->edns->size($UDPmax) if $query->edns->defined;
+	return $reply unless grep $_->isa('Net::DNS::RR::OPT'), @{$query->{additional}};
+
+	my $edns = $reply->edns();
+	CORE::push( @{$reply->{additional}}, $edns );
+	$edns->size($UDPmax);
 	return $reply;
 }
 
@@ -386,7 +384,8 @@ sub string {
 	my $update = $header->opcode eq 'UPDATE';
 
 	my $server = $self->{answerfrom};
-	my $string = $server ? ";; Answer received from $server ($self->{answersize} bytes)\n" : "";
+	my $length = $self->{answersize};
+	my $string = $server ? ";; Answer received from $server ($length bytes)\n" : "";
 
 	$string .= ";; HEADER SECTION\n" . $header->string;
 
@@ -429,9 +428,8 @@ User-created packets will return undef for this method.
 sub answerfrom {
 	my $self = shift;
 
-	return $self->{answerfrom} = shift if scalar @_;
-
-	return $self->{answerfrom};
+	$self->{answerfrom} = shift if scalar @_;
+	$self->{answerfrom};
 }
 
 
@@ -446,7 +444,7 @@ nameserver.  User-created packets will return undef for this method
 =cut
 
 sub answersize {
-	return shift->{answersize};
+	shift->{answersize};
 }
 
 
@@ -463,21 +461,14 @@ Adds RRs to the specified section of the packet.
 
 Returns the number of resource records in the specified section.
 
+Section names may be abbreviated to the first three characters.
+
 =cut
 
 sub push {
 	my $self = shift;
 	my $list = $self->_section(shift);
-	my @rr	 = grep ref($_), @_;
-
-	if ( $self->header->opcode eq 'UPDATE' ) {
-		my $zclass = ( $self->zone )[0]->zclass;
-		foreach (@rr) {
-			$_->class($zclass) unless $_->class =~ /ANY|NONE/;
-		}
-	}
-
-	return CORE::push( @$list, @rr );
+	CORE::push( @$list, grep ref($_), @_ );
 }
 
 
@@ -495,6 +486,8 @@ RRs are not already present in the same section.
 
 Returns the number of resource records in the specified section.
 
+Section names may be abbreviated to the first three characters.
+
 =cut
 
 sub unique_push {
@@ -502,22 +495,9 @@ sub unique_push {
 	my $list = $self->_section(shift);
 	my @rr	 = grep ref($_), @_;
 
-	if ( $self->header->opcode eq 'UPDATE' ) {
-		my $zclass = ( $self->zone )[0]->zclass;
-		foreach (@rr) {
-			$_->class($zclass) unless $_->class =~ /ANY|NONE/;
-		}
-	}
+	my %unique = map { ( bless( {%$_, ttl => 0}, ref $_ )->canonical => $_ ) } @rr, @$list;
 
-	my %unique = map { ( bless( {%$_, ttl => 0}, ref($_) )->canonical, $_ ) } @$list, @rr;
-
-	@$list = ();
-	return CORE::push( @$list, values %unique );
-}
-
-sub safe_push {
-	carp 'safe_push() deprecated: replaced by unique_push()';
-	&unique_push;
+	scalar( @$list = values %unique );
 }
 
 
@@ -534,8 +514,7 @@ Removes a single RR from the specified section of the packet.
 sub pop {
 	my $self = shift;
 	my $list = $self->_section(shift);
-
-	return CORE::pop(@$list);
+	CORE::pop(@$list);
 }
 
 
@@ -551,104 +530,7 @@ sub _section {				## returns array reference for section
 	my $self = shift;
 	my $name = shift;
 	my $list = $_section{unpack 'a3', $name} || $name;
-	return $self->{$list} || [];
-}
-
-
-# =head2 dn_comp
-
-#     $compname = $packet->dn_comp("foo.example.com", $offset);
-
-# Returns a domain name compressed for a particular packet object, to
-# be stored beginning at the given offset within the packet data.  The
-# name will be added to a running list of compressed domain names for
-# future use.
-
-# =cut
-
-sub dn_comp {
-	my ($self, $fqdn, $offset) = @_;
-
-	my @labels = Net::DNS::name2labels($fqdn);
-	my $hash   = $self->{compnames};
-	my $data   = '';
-	while (@labels) {
-		my $name = join( '.', @labels );
-
-		return $data . pack( 'n', 0xC000 | $hash->{$name} ) if defined $hash->{$name};
-
-		my $label = shift @labels;
-		my $length = length($label) || next;		   # skip if null
-		if ( $length > 63 ) {
-			$length = 63;
-			$label = substr( $label, 0, $length );
-			carp "\n$label...\ntruncated to $length octets (RFC1035 2.3.1)";
-		}
-		$data .= pack( 'C a*', $length, $label );
-
-		next unless $offset < 0x4000;
-		$hash->{$name} = $offset;
-		$offset += 1 + $length;
-	}
-	$data .= chr(0);
-}
-
-
-# =head2 dn_expand
-
-#     use Net::DNS::Packet qw(dn_expand);
-#     ($name, $nextoffset) = dn_expand(\$data, $offset);
-
-#     ($name, $nextoffset) = Net::DNS::Packet::dn_expand(\$data, $offset);
-
-# Expands the domain name stored at a particular location in a DNS
-# packet.  The first argument is a reference to a scalar containing
-# the packet data.  The second argument is the offset within the
-# packet where the (possibly compressed) domain name is stored.
-
-# Returns the domain name and the offset of the next location in the
-# packet.
-
-# Returns undef if the domain name could not be expanded.
-
-# =cut
-
-
-# This is very hot code, so we try to keep things fast.  This makes for
-# odd style sometimes.
-
-sub dn_expand {
-#FYI	my ($packet, $offset) = @_;
-	return dn_expand_XS(@_) if $Net::DNS::HAVE_XS;
-#	warn "USING PURE PERL dn_expand()\n";
-	return dn_expand_PP(@_, {} );	# $packet, $offset, anonymous hash
-}
-
-sub dn_expand_PP {
-	my ($packet, $offset, $visited) = @_;
-	my $packetlen = length $$packet;
-	my $name = '';
-
-	while ( $offset < $packetlen ) {
-		unless ( my $length = unpack("\@$offset C", $$packet) ) {
-			$name =~ s/\.$//o;
-			return ($name, ++$offset);
-
-		} elsif ( ($length & 0xc0) == 0xc0 ) {		# pointer
-			my $point = 0x3fff & unpack("\@$offset n", $$packet);
-			die 'Exception: unbounded name expansion' if $visited->{$point}++;
-
-			my ($suffix) = dn_expand_PP($packet, $point, $visited);
-
-			return ($name.$suffix, $offset+2) if defined $suffix;
-
-		} else {
-			my $element = substr($$packet, ++$offset, $length);
-			$name .= Net::DNS::wire2presentation($element).'.';
-			$offset += $length;
-		}
-	}
-	return undef;
+	$self->{$list} ||= [];
 }
 
 
@@ -723,13 +605,15 @@ does not support the suppressed signature scheme described in RFC2845.
 sub sign_tsig {
 	my $self = shift;
 
-	my $tsig = eval {
+	eval {
+		local $SIG{__DIE__};
 		require Net::DNS::RR::TSIG;
-		Net::DNS::RR::TSIG->create(@_);
-	} || croak "$@\nTSIG: unable to sign packet";
-
-	$self->push( 'additional', $tsig );
-	return $tsig;
+		my $tsig = Net::DNS::RR::TSIG->create(@_);
+		$self->push( 'additional' => $tsig );
+		return $tsig;
+	} || do {
+		croak "$@\nTSIG: unable to sign packet";
+	};
 }
 
 
@@ -756,29 +640,30 @@ on the final packet in the absence of more specific information.
 sub verify {
 	my $self = shift;
 
-	my $sig = $self->sigrr || return undef;
-	return $sig->verify( $self, @_ );
+	my $sig = $self->sigrr;
+	return $sig ? $sig->verify( $self, @_ ) : shift;
 }
 
 sub verifyerr {
 	my $self = shift;
 
-	my $sig = $self->sigrr || return 'not signed';
-	return $sig->vrfyerrstr;
+	my $sig = $self->sigrr;
+	return $sig ? $sig->vrfyerrstr : 'not signed';
 }
 
 
 =head2 sign_sig0
 
 SIG0 support is provided through the Net::DNS::RR::SIG class.
-This class is not integrated into Net::DNS but resides in the
-Net::DNS::SEC distribution available from CPAN.
+The requisite cryptographic components are not integrated into
+Net::DNS but reside in the Net::DNS::SEC distribution available
+from CPAN.
 
     $update = new Net::DNS::Update('example.com');
     $update->push( update => rr_add('foo.example.com A 10.1.2.3'));
     $update->sign_sig0('Kexample.com+003+25317.private');
 
-Execution will be terminated if Net::DNS::RR::SIG is not available.
+Execution will be terminated if Net::DNS::SEC is not available.
 
 
 =head2 verify SIG0
@@ -792,27 +677,50 @@ Verify SIG0 packet signature against one or more specified KEY RRs.
 
 sub sign_sig0 {
 	my $self = shift;
-	my $karg = shift || return undef;
+	my $karg = shift;
 
-	my $sig0 = eval {
-		unless ( my $kref = ref($karg) ) {
-			return Net::DNS::RR::SIG->create( '', $karg );
+	eval {
+		local $SIG{__DIE__};
+		require Net::DNS::RR::SIG;
 
-		} elsif ( $kref eq 'Net::DNS::RR::SIG' ) {
-			return $karg;
+		my $sig0;
+		if ( ref($karg) eq 'Net::DNS::RR::SIG' ) {
+			$sig0 = $karg;
 
-		} elsif ( $kref eq 'Net::DNS::SEC::Private' ) {
-			return Net::DNS::RR::SIG->create( '', $karg );
 		} else {
-			die "unexpected $kref argument passed to sign_sig0";
+			$sig0 = Net::DNS::RR::SIG->create( '', $karg );
 		}
-	} || croak "$@\nSIG0: Net::DNS::SEC not available";
 
-	$self->push( 'additional', $sig0 );
-	return $sig0;
+		$self->push( 'additional' => $sig0 );
+		return $sig0;
+	} || do {
+		croak "$@\nSIG0: unable to sign packet";
+	};
 }
 
 
+=head2 sigrr
+
+    $sigrr = $packet->sigrr() || die 'unsigned packet';
+
+The sigrr method returns the signature RR from a signed packet
+or undefined if the signature is absent.
+
+=cut
+
+sub sigrr {
+	my $self = shift;
+
+	my ($sig) = reverse $self->additional;
+	return undef unless $sig;
+	return $sig if $sig->type eq 'TSIG';
+	return $sig if $sig->type eq 'SIG';
+	return undef;
+}
+
+
+########################################
+
 =head2 truncate
 
 The truncate method takes a maximum length as argument and then tries
@@ -824,7 +732,8 @@ The minimum maximum length that is honoured is 512 octets.
 =cut
 
 # From RFC2181:
-#9. The TC (truncated) header bit
+#
+# 9. The TC (truncated) header bit
 #
 #   The TC bit should be set in responses only when an RRSet is required
 #   as a part of the response, but could not be included in its entirety.
@@ -841,108 +750,114 @@ The minimum maximum length that is honoured is 512 octets.
 #   set, it should ignore that response, and query again, using a
 #   mechanism, such as a TCP connection, that will permit larger replies.
 
-# Code inspired on a contribution from Aaron Crane via rt.cpan.org 33547
+# Code developed from a contribution by Aaron Crane via rt.cpan.org 33547
 
 sub truncate {
-	my $self=shift;
-	my $max_len=shift;
-	my $debug=0;
-	$max_len=$max_len>UDPSZ?$max_len:UDPSZ;
+	my $self = shift;
+	my $size = shift || UDPSZ;
 
-	print "Truncating to $max_len\n" if $debug;
+	my $sigrr = $self->sigrr;
+	$size = UDPSZ unless $size > UDPSZ;
+	$size -= $sigrr->_size if $sigrr;
 
-	if (length $self->encode() > $max_len) {
-		# first remove data from the additional section
-		while (length $self->encode() > $max_len){
-			# first remove _complete_ RRstes from the additonal section.
-			my $popped= CORE::pop(@{$self->{'additional'}});
-			last unless defined($popped);
-			print "Removed ".$popped->string." from additional \n" if $debug;
-			my $i=0;
-			my @stripped_additonal;
+	my $data = pack 'x' x 12;				# header placeholder
+	my $hdsz = length $data;
+	$self->{count} = [];
 
-			while ( $i < scalar @{$self->{'additional'}} ) {
-				#remove all of these same RRtypes
-				if  (
-				    ${$self->{'additional'}}[$i]->type eq $popped->type &&
-				    ${$self->{'additional'}}[$i]->name eq $popped->name &&
-				    ${$self->{'additional'}}[$i]->class eq $popped->class ){
-					print "       Also removed ". ${$self->{'additional'}}[$i]->string." from additonal \n" if $debug;				}else{
-					CORE::push @stripped_additonal,  ${$self->{'additional'}}[$i];
-				}
-				$i++;
-			}
-			$self->{'additional'}=\@stripped_additonal;
-		}
-
-		return $self if length $self->encode <= $max_len;
-
-      		my @sections = qw<authority answer question>;
-		while (@sections) {
-			while (my $popped=$self->pop($sections[0])) {
-				last unless defined($popped);
-				print "Popped ".$popped->string." from the $sections[0] section\n" if $debug;
-				$self->header->tc(1);
-				return $self if length $self->encode <= $max_len;
-				next;
-			}
-			shift @sections;
+	my $tc;
+	my $hash = {};
+	foreach my $section ( map $self->{$_}, qw(question answer authority) ) {
+		my @list;
+		foreach my $item (@$section) {
+			my $component = $item->encode( length $data, $hash );
+			last if length($data) + length($component) > $size;
+			last if $tc;
+			$data .= $component;
+			CORE::push @list, $item;
 		}
+		$tc++ if scalar(@list) < scalar(@$section);
+		@$section = @list;
 	}
-	return $self;
+	$self->header->tc(1) if $tc;				# only set if truncated here
+
+	my %rrset;
+	my @order;
+	foreach my $item ( grep ref($_) ne ref($sigrr), $self->additional ) {
+		my $name  = $item->{owner}->canonical;
+		my $class = $item->{class} || 0;
+		my $key	  = pack 'nna*', $class, $item->{type}, $name;
+		CORE::push @order, $key unless $rrset{$key};
+		CORE::push @{$rrset{$key}}, $item;
+	}
+
+	my @list;
+	foreach my $key (@order) {
+		my $component = '';
+		my @item      = @{$rrset{$key}};
+		foreach my $item (@item) {
+			$component .= $item->encode( length $data, $hash );
+		}
+		last if length($data) + length($component) > $size;
+		$data .= $component;
+		CORE::push @list, @item;
+	}
+
+	if ($sigrr) {
+		$data .= $sigrr->encode( length $data, $hash, $self );
+		CORE::push @list, $sigrr;
+	}
+	$self->{'additional'} = \@list;
+
+	my @part = qw(question answer authority additional);
+	my @size = map scalar( @{$self->{$_}} ), @part;
+	pack 'n6 a*', $self->header->id, $self->{status}, @size, substr( $data, $hdsz );
 }
 
 
 ########################################
 
 sub dump {				## print internal data structure
-	require Data::Dumper;
-	local $Data::Dumper::Maxdepth = 6;
-	local $Data::Dumper::Sortkeys = 1;
+	require Data::Dumper;					# uncoverable pod
+	local $Data::Dumper::Maxdepth = $Data::Dumper::Maxdepth || 3;
+	local $Data::Dumper::Sortkeys = $Data::Dumper::Sortkeys || 1;
 	print Data::Dumper::Dumper(@_);
 }
 
 
-sub sigrr {				## obtain packet signature RR
-	my $self = shift;
-
-	my ($sig) = reverse $self->additional;
-	return undef unless $sig;
-	return $sig if $sig->type eq 'TSIG';
-	return $sig if $sig->type eq 'SIG';
-	return undef;
-}
-
-
-sub DESTROY { }				## Avoid tickling AUTOLOAD (in cleanup)
-
-use vars qw($AUTOLOAD);
-
-sub AUTOLOAD {				## Default method
-	no strict;
-	@_ = ("method $AUTOLOAD undefined");
-	goto &{'Carp::confess'};
-}
-
-
 1;
 __END__
 
 
 =head1 COPYRIGHT
 
-Copyright (c)1997-2002 Michael Fuhr.
+Copyright (c)1997-2000 Michael Fuhr.
 
 Portions Copyright (c)2002-2004 Chris Reinhardt.
 
 Portions Copyright (c)2002-2009 Olaf Kolkman
 
-Portions Copyright (c)2007-2013 Dick Franks
+Portions Copyright (c)2007-2015 Dick Franks
 
 All rights reserved.
 
-This program is free software; you may redistribute it and/or
-modify it under the same terms as Perl itself.
+
+=head1 LICENSE
+
+Permission to use, copy, modify, and distribute this software and its
+documentation for any purpose and without fee is hereby granted, provided
+that the above copyright notice appear in all copies and that both that
+copyright notice and this permission notice appear in supporting
+documentation, and that the name of the author not be used in advertising
+or publicity pertaining to distribution of the software without specific
+prior written permission.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
+THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+DEALINGS IN THE SOFTWARE.
 
 
 =head1 SEE ALSO
